#define MAX_HOSTLEN 16
#define MAX_DOMLEN 16
#define MAX_USERLEN 16
#define RESP_LEN 24
#define NONCE_LEN 8

struct ntlm_msg1 {
	unsigned char protocol[8];
	unsigned char type; /* 1 */
	unsigned char zero1[3];
	unsigned char flags1;
	unsigned char flags2;
	unsigned char zero2[2];
	
        unsigned char dom_len[4];
	unsigned char dom_off[4];
	
	unsigned char host_len[4];
	unsigned char host_off[4];
	
	unsigned char data[0];
} __attribute__((packed));

struct ntlm_msg2 {
	unsigned char protocol[8];
	unsigned char type; /* 2 */
	unsigned char zero1[7];
	unsigned char msg_len[4];
	unsigned char flags[2];
	unsigned char zero2[2];
	
	unsigned char nonce[8];
	unsigned char zero3[8];
} __attribute__((packed));

struct ntlm_msg3 {
	unsigned char protocol[8];
	unsigned char type; /* 3 */
	unsigned char zero1[3];
	
	unsigned char lm_len[4];
	unsigned char lm_off[4];
	
	unsigned char nt_len[4];
	unsigned char nt_off[4];
	
	unsigned char dom_len[4];
	unsigned char dom_off[4];
	
	unsigned char user_len[4];
	unsigned char user_off[4];
	
	unsigned char host_len[4];
	unsigned char host_off[4];
	
	unsigned char msg_len[4];

        unsigned char data[0];
} __attribute__((packed));

typedef struct ntlmssp_info {
        int msg_type;
        unsigned char user[MAX_USERLEN+1];
        unsigned char host[MAX_HOSTLEN+1];
        unsigned char domain[MAX_DOMLEN+1];
        unsigned char lm[RESP_LEN];
        unsigned char nt[RESP_LEN];
} ntlmssp_info_rec;

#define little_endian_word(x) x[0] + (((unsigned)x[1]) << 8)
#define set_little_endian_word(x,y) x[0]=(y&0xff);x[1]=(y>>8)

static int ntlm_msg_type(unsigned char *raw_msg, unsigned msglen)
{
        struct ntlm_msg1 *msg = (struct ntlm_msg1 *)raw_msg;

        if (msglen < 9) return -1;
        if (strncmp(msg->protocol, "NTLMSSP", 8))
                return -1;
        return msg->type;
}

static int ntlm_extract_mem(unsigned char *dst, unsigned char *src,
unsigned srclen, unsigned char *off, unsigned char *len, unsigned max)
{
	unsigned o = little_endian_word(off);
	unsigned l = little_endian_word(len);
	if (l > max) return -1;
        if (o >= srclen) return -1;
        if (o+l > srclen) return -1;
        src += o;
        while (l-- > 0)
                *dst++ = *src++;
        return 0;
}

static int ntlm_extract_string(unsigned char *dst, unsigned char *src,
unsigned srclen, unsigned char *off, unsigned char *len, unsigned max)
{
	unsigned o = little_endian_word(off);
	unsigned l = little_endian_word(len);
	if (l > max) return -1;
        if (o >= srclen) return -1;
        if (o+l > srclen) return -1;
        src += o;
        while (l-- > 0)
                *dst++ = *src++;
        *dst = 0;
        return 0;
}

static int ntlm_extract_unicode(unsigned char *dst, unsigned char *src,
unsigned srclen, unsigned char *off, unsigned char *len, unsigned max)
{
	unsigned o = little_endian_word(off);
	unsigned l = little_endian_word(len) / 2; /* Unicode! */
	if (l > max) return -1;
        if (o >= srclen) return -1;
        if (o+l > srclen) return -1;
        src += o;
        while (l > 0) {
                /* Unicode to ASCII */
                *dst++ = *src;
                src += 2;
                l -= 2;
        }
        *dst = 0;
        return 0;
}

static int ntlm_msg1_gethostname(unsigned char *raw_msg, unsigned msglen,
unsigned char *hostname)
{
        struct ntlm_msg1 *msg = (struct ntlm_msg1 *)raw_msg;
	return ntlm_extract_string(hostname, (char *)msg, msglen, msg->host_off, msg->host_len, MAX_HOSTLEN);
}

static int ntlm_msg1_getdomainname(unsigned char *raw_msg, unsigned msglen,
unsigned char *domainname)
{
        struct ntlm_msg1 *msg = (struct ntlm_msg1 *)raw_msg;
	return ntlm_extract_string(domainname, (char *)msg, msglen, msg->dom_off, msg->dom_len, MAX_DOMLEN);
}

static int ntlm_msg3_getlm(unsigned char *raw_msg, unsigned msglen,
unsigned char *lm)
{
        struct ntlm_msg3 *msg = (struct ntlm_msg3 *)raw_msg;
        return ntlm_extract_mem(lm, (char *)msg, msglen, msg->lm_off, msg->lm_len, RESP_LEN);
}

static int ntlm_msg3_getnt(unsigned char *raw_msg, unsigned msglen,
unsigned char *nt)
{
        struct ntlm_msg3 *msg = (struct ntlm_msg3 *)raw_msg;
        return ntlm_extract_mem(nt, (char *)msg, msglen, msg->nt_off, msg->nt_len, RESP_LEN);
}

static int ntlm_msg3_getusername(unsigned char *raw_msg,
unsigned msglen, unsigned char *username)
{
        struct ntlm_msg3 *msg = (struct ntlm_msg3 *)raw_msg;
	return ntlm_extract_unicode(username, (char *)msg, msglen, msg->user_off, msg->user_len, MAX_USERLEN);
}

static int ntlm_msg3_gethostname(unsigned char *raw_msg, unsigned msglen,
unsigned char *hostname)
{
        struct ntlm_msg3 *msg = (struct ntlm_msg3 *)raw_msg;
	return ntlm_extract_unicode(hostname, (char *)msg, msglen, msg->host_off, msg->host_len, MAX_HOSTLEN);
}

static int ntlm_msg3_getdomainname(unsigned char *raw_msg, unsigned msglen,
unsigned char *domainname)
{
        struct ntlm_msg3 *msg = (struct ntlm_msg3 *)raw_msg;
	return ntlm_extract_unicode(domainname, (char *)msg, msglen, msg->dom_off, msg->dom_len, MAX_DOMLEN);
}

static int ntlm_decode_msg(struct ntlmssp_info *info,
unsigned char *raw_msg, unsigned msglen)
{
        switch (info->msg_type = ntlm_msg_type(raw_msg, msglen)) {
                case 1: return ntlm_msg1_gethostname(raw_msg, msglen, info->host)
                        + ntlm_msg1_getdomainname(raw_msg, msglen, info->domain);
                case 3: return ntlm_msg3_getlm(raw_msg, msglen, info->lm)
                        + ntlm_msg3_getnt(raw_msg, msglen, info->nt)
                        + ntlm_msg3_getusername(raw_msg, msglen, info->user)
                        + ntlm_msg3_gethostname(raw_msg, msglen, info->host)
                        + ntlm_msg3_getdomainname(raw_msg, msglen, info->domain);
        }
        return -1;
}

static int ntlm_encode_msg2(unsigned char *nonce, struct ntlm_msg2 *msg)
{
        memset(msg, 0, sizeof(struct ntlm_msg2));
        strcpy(msg->protocol, "NTLMSSP");
        msg->type = 0x02;
        set_little_endian_word(msg->msg_len, sizeof(struct ntlm_msg2));
        set_little_endian_word(msg->flags, 0x8201);
        memcpy(msg->nonce, nonce, sizeof(msg->nonce));
        return 0;
}

